/*
 * Unit test suite for clipboard functions.
 *
 * Copyright 2002 Dmitry Timoshkov
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "wine/test.h"
#include "winbase.h"
#include "winerror.h"
#include "winuser.h"

static void test_ClipboardOwner(void)
{
    HWND hWnd1, hWnd2;

    SetLastError(0);
    ok(!GetClipboardOwner() && GetLastError() == 0,
       "could not perform clipboard test: clipboard already owned");

    hWnd1 = CreateWindowExA(0, "static", NULL, WS_POPUP,
                                 0, 0, 10, 10, 0, 0, 0, NULL);
    ok(hWnd1 != 0, "CreateWindowExA error %ld", GetLastError());
    trace("hWnd1 = %p\n", hWnd1);

    hWnd2 = CreateWindowExA(0, "static", NULL, WS_POPUP,
                                 0, 0, 10, 10, 0, 0, 0, NULL);
    ok(hWnd2 != 0, "CreateWindowExA error %ld", GetLastError());
    trace("hWnd2 = %p\n", hWnd2);

    ok(!CloseClipboard() && GetLastError() == ERROR_CLIPBOARD_NOT_OPEN,
       "CloseClipboard should fail with error ERROR_CLIPBOARD_NOT_OPEN if clipboard wasn't open");

    ok(OpenClipboard(0), "OpenClipboard failed");
    ok(!GetClipboardOwner(), "clipboard should still be not owned");
    ok(!OpenClipboard(hWnd1), "OpenClipboard should fail since clipboard already opened");
    ok(CloseClipboard(), "CloseClipboard error %ld", GetLastError());

    ok(OpenClipboard(hWnd1), "OpenClipboard failed");

    SetLastError(0);
    ok(!OpenClipboard(hWnd2) && GetLastError() == 0,
       "OpenClipboard should fail without setting last error value");

    SetLastError(0);
    ok(!GetClipboardOwner() && GetLastError() == 0, "clipboard should still be not owned");
    ok(EmptyClipboard(), "EmptyClipboard error %ld", GetLastError());
    ok(GetClipboardOwner() == hWnd1, "clipboard should be owned by %p", hWnd1);

    SetLastError(0);
    ok(!OpenClipboard(hWnd2) && GetLastError() == 0,
       "OpenClipboard should fail without setting last error value");

    ok(CloseClipboard(), "CloseClipboard error %ld", GetLastError());
    ok(GetClipboardOwner() == hWnd1, "clipboard should still be owned");

    ok(DestroyWindow(hWnd1), "DestroyWindow error %ld", GetLastError());
    ok(DestroyWindow(hWnd2), "DestroyWindow error %ld", GetLastError());
    SetLastError(0);
    ok(!GetClipboardOwner() && GetLastError() == 0, "clipboard should not be owned");
}

static void test_RegisterClipboardFormatA(void)
{
    ATOM atom_id;
    UINT format_id, format_id2;
    char buf[256];
    int len;

    format_id = RegisterClipboardFormatA("my_cool_clipboard_format");
    ok(format_id > 0xc000 && format_id < 0xffff, "invalid clipboard format id %04x", format_id);

    format_id2 = RegisterClipboardFormatA("MY_COOL_CLIPBOARD_FORMAT");
    ok(format_id2 == format_id, "invalid clipboard format id %04x", format_id2);

    len = GetClipboardFormatNameA(format_id, buf, 256);
    ok(len == lstrlenA("my_cool_clipboard_format"), "wrong format name length %d", len);
    ok(!lstrcmpA(buf, "my_cool_clipboard_format"),  "wrong format name");

    lstrcpyA(buf, "foo");
    len = GetAtomNameA((ATOM)format_id, buf, 256);
    ok(len == 0 && GetLastError() == ERROR_INVALID_HANDLE,
       "GetAtomNameA should fail with error ERROR_INVALID_HANDLE");
    ok(!lstrcmpA(buf, "foo"),  "wrong format name");

todo_wine
{
    lstrcpyA(buf, "foo");
    len = GlobalGetAtomNameA((ATOM)format_id, buf, 256);
    ok(len == 0 && GetLastError() == ERROR_INVALID_HANDLE,
       "GlobalGetAtomNameA should fail with error ERROR_INVALID_HANDLE");
    ok(!lstrcmpA(buf, "foo"),  "wrong format name");
}

    atom_id = FindAtomA("my_cool_clipboard_format");
    ok(atom_id == 0 && GetLastError() == ERROR_FILE_NOT_FOUND,
       "FindAtomA should fail with error ERROR_FILE_NOT_FOUND");

#if 0
    /* this relies on the clipboard and global atom table being different */
    atom_id = GlobalFindAtomA("my_cool_clipboard_format");
    ok(atom_id == 0 && GetLastError() == ERROR_FILE_NOT_FOUND,
       "GlobalFindAtomA should fail with error ERROR_FILE_NOT_FOUND");

    for (format_id = 0; format_id < 0xffff; format_id++)
    {
        SetLastError(0);
        len = GetClipboardFormatNameA(format_id, buf, 256);

        if (format_id < 0xc000)
        {
            ok(!len && GetLastError() == ERROR_INVALID_PARAMETER,
               "GetClipboardFormatName should fail with ERROR_INVALID_PARAMETER for predefined clipboard format");
        }
if (0) ;/* commented out above */
        else
        {
            if (len)
                trace("%04x: %s\n", format_id, len ? buf : "");
            else
                ok(GetLastError() == ERROR_INVALID_HANDLE,
                   "GetClipboardFormatName should fail with ERROR_INVALID_HANDLE for not predefined clipboard format");
        }
    }
#endif

    ok(OpenClipboard(0), "OpenClipboard error %ld", GetLastError());

    trace("# of formats available: %d\n", CountClipboardFormats());

    format_id = 0;
    while ((format_id = EnumClipboardFormats(format_id)))
    {
        ok(IsClipboardFormatAvailable(format_id), "format %04x was listed as available", format_id);
        len = GetClipboardFormatNameA(format_id, buf, 256);
        trace("%04x: %s\n", format_id, len ? buf : "");
    }

    ok(EmptyClipboard(), "EmptyClipboard error %ld", GetLastError());
    ok(CloseClipboard(), "CloseClipboard error %ld", GetLastError());

    if (CountClipboardFormats())
        ok(!EnumClipboardFormats(0) && GetLastError() == ERROR_CLIPBOARD_NOT_OPEN,
           "EnumClipboardFormats should fail with error ERROR_CLIPBOARD_NOT_OPEN if clipboard wasn't open");

    ok(!EmptyClipboard() && GetLastError() == ERROR_CLIPBOARD_NOT_OPEN,
       "EmptyClipboard should fail with error ERROR_CLIPBOARD_NOT_OPEN if clipboard wasn't open");
}

START_TEST(clipboard)
{
    test_RegisterClipboardFormatA();
    test_ClipboardOwner();
}
